/*
 *  This is a sample plugin module
 *
 *  It can be compiled by any of the supported compilers:
 *
 *      - Borland C++, CBuilder, free C++
 *      - Watcom C++ for DOS32
 *      - Watcom C++ for OS/2
 *      - Visual C++
 *
 */

#include <ida.hpp>
#include <idp.hpp>
#include <bytes.hpp>
#include <loader.hpp>
#include <kernwin.hpp>
#include <name.hpp>
#include "md5.h"
#include <string.h>

static void load_fnbase(const char* x);
static void resolveFunctions();

//--------------------------------------------------------------------------
// This callback is called for UI notification events
static int sample_callback(void * /*user_data*/, int event_id, va_list /*va*/)
{
  if ( event_id != ui_msg )     // avoid recursion
    if ( event_id != ui_setstate
      && event_id != ui_showauto
      && event_id != ui_refreshmarked ) // ignore uninteresting events
                    msg("ui_callback %d\n", event_id);
  return 0;                     // 0 means "process the event"
                                // otherwise the event would be ignored
}

//--------------------------------------------------------------------------
// A sample how to generate user-defined line prefixes
static const int prefix_width = 8;

static void get_user_defined_prefix(ea_t ea,
                                    int lnnum,
                                    int indent,
                                    const char *line,
                                    char *buf,
                                    size_t bufsize)
{
  buf[0] = '\0';        // empty prefix by default

  // We want to display the prefix only the lines which
  // contain the instruction itself

  if ( indent != -1 ) return;           // a directive
  if ( line[0] == '\0' ) return;        // empty line
  if ( *line == COLOR_ON ) line += 2;
  if ( *line == ash.cmnt[0] ) return;   // comment line...

  // We don't want the prefix to be printed again for other lines of the
  // same instruction/data. For that we remember the line number
  // and compare it before generating the prefix

  static ea_t old_ea = BADADDR;
  static int old_lnnum;
  if ( old_ea == ea && old_lnnum == lnnum ) return;

  // Let's display the size of the current item as the user-defined prefix
  ulong our_size = get_item_size(ea);

  // seems to be an instruction line. we don't bother about the width
  // because it will be padded with spaces by the kernel

  snprintf(buf, bufsize, " %d", our_size);

  // Remember the address and line number we produced the line prefix for:
  old_ea = ea;
  old_lnnum = lnnum;

}

//--------------------------------------------------------------------------
//
//      Initialize.
//
//      IDA will call this function only once.
//      If this function returns PLGUIN_SKIP, IDA will never load it again.
//      If this function returns PLUGIN_OK, IDA will unload the plugin but
//      remember that the plugin agreed to work with the database.
//      The plugin will be loaded again if the user invokes it by
//      pressing the hotkey or selecting it from the menu.
//      After the second load the plugin will stay on memory.
//      If this function returns PLUGIN_KEEP, IDA will keep the plugin
//      in the memory. In this case the initialization function can hook
//      into the processor module and user interface notification points.
//      See the hook_to_notification_point() function.
//
//      In this example we check the input file format and make the decision.
//      You may or may not check any other conditions to decide what you do:
//      whether you agree to work with the database or not.
//
int init(void)
{
  if ( inf.filetype == f_ELF ) return PLUGIN_KEEP;

// Please uncomment the following line to see how the notification works
//  hook_to_notification_point(HT_UI, sample_callback, NULL);

// Please uncomment the following line to see how the user-defined prefix works
//  set_user_defined_prefix(prefix_width, get_user_defined_prefix);

  return PLUGIN_SKIP;
}

//--------------------------------------------------------------------------
//      Terminate.
//      Usually this callback is empty.
//      The plugin should unhook from the notification lists if
//      hook_to_notification_point() was used.
//
//      IDA will call this function when the user asks to exit.
//      This function won't be called in the case of emergency exits.

void term(void)
{
  unhook_from_notification_point(HT_UI, sample_callback);
  set_user_defined_prefix(0, NULL);
}

//--------------------------------------------------------------------------
//
//      The plugin method
//
//      This is the main function of plugin.
//
//      It will be called when the user selects the plugin.
//
//              arg - the input argument, it can be specified in
//                    plugins.cfg file. The default is zero.
//
//

void run(int arg)
{
	char *fileName;

	fileName = askfile_c(0, "*.dat", "Please choose libc db file");

	if ( fileName != NULL ){
		msg(fileName);
	}

	load_fnbase(fileName);
	resolveFunctions();
}

//--------------------------------------------------------------------------
char comment[] = "This is a sample plugin. It does nothing useful";

char help[] =
        "A sample plugin module\n"
        "\n"
        "This module shows you how to create plugin modules.\n"
        "\n"
        "It does nothing useful - just prints a message that is was called\n"
        "and shows the current address.\n";


//--------------------------------------------------------------------------
// This is the preferred name of the plugin module in the menu system
// The preferred name may be overriden in plugins.cfg file

char wanted_name[] = "Sample plugin";


// This is the preferred hotkey for the plugin module
// The preferred hotkey may be overriden in plugins.cfg file
// Note: IDA won't tell you if the hotkey is not correct
//       It will just disable the hotkey.

char wanted_hotkey[] = "Alt-0";


//--------------------------------------------------------------------------
//
//      PLUGIN DESCRIPTION BLOCK
//
//--------------------------------------------------------------------------

extern "C" plugin_t PLUGIN = {
  IDP_INTERFACE_VERSION,
  0,                    // plugin flags
  init,                 // initialize

  term,                 // terminate. this pointer may be NULL.

  run,                  // invoke plugin

  comment,              // long comment about the plugin
                        // it could appear in the status line
                        // or as a hint

  help,                 // multiline help about the plugin

  wanted_name,          // the preferred short name of the plugin
  wanted_hotkey         // the preferred hotkey to run the plugin
};


#define SIGNATSIZE       24     // Size of function signature

static char * find_fprints(unsigned int a, int *);
static inline char * handle_fncall(unsigned char sig[SIGNATSIZE+4], int *);


struct fenris_fndb {
  char* name;
  unsigned short a;
  struct fenris_fndb* next;
};

struct fenris_fndb*      fndb[256*256]; // Speed search table


										
static void resolveFunctions(){
	int qty, i, j;
	char funcName[64];
	unsigned char sig[SIGNATSIZE+4];
	char *result;
	char *newName;
	int count;

	qty = get_func_qty();
	msg("Total number of functions: %d\n", qty);
	for (i = 0; i < qty; i++){
		func_t * cur_func = getn_func(i);
		result = get_name(cur_func->startEA, cur_func->startEA, funcName, 64);
		if ( result != NULL ){
			msg("%s", result);
		}
		
		if ( strncmp(result, "sub_", 4) == 0 ){
			for (j = 0; j < SIGNATSIZE+4; j++){
				sig[j] = get_byte(cur_func->startEA + j);
			}
			
			newName = handle_fncall(sig, &count);
			if (count == 1){
				char *response;

				response = askstr(HIST_IDENT, newName, "Do you want to rename %s with %s?", result, newName);
				if ( response != NULL ){
					set_name(cur_func->startEA, strdup(response));
				}
			}
			else if ( count > 1 ){
				char *response = askstr(HIST_IDENT, newName, "There are multiple matches. Do you want to rename %s with %s?", result, newName);
				if ( response != NULL ){
					set_name(cur_func->startEA, strdup(response));
				}
			}
		}
	}
}
										
/******************************
 * Load fingerprints database *
 ******************************/

static void load_fnbase(const char* x) {
  char buf[500];
  FILE* f;
  unsigned int a;

  f=fopen(x,"r");


  if (!f) {
    warning("* WARNING: cannot load '%s' fingerprints database.",x);
    return;
  }

  show_wait_box("Please wait while the fingerprints database is being loaded...");
  // Now, this is going to be a bit awkward, but gives us great
  // benefit later, when searching. Use a table to lookup first two
  // bytes instantly, then compare two remaining bytes by searching
  // a linked list.

  while (fgets(buf,sizeof(buf)-1,f)) {
    struct fenris_fndb ff;
    char* x,*fname;
    x=strchr(buf,' ');
    if (!x) continue; // Doh?
    fname=x+1;
    x=strchr(x+1,' ');
    if (!x) continue; // Doh?
    *x=0;
    if (sscanf(x+1,"%X",&a)!=1) continue; // Doh?!

    ff.a=a & 0xffff;
    ff.name=strdup(fname);
    ff.next=0;

    if (!fndb[a>>16]) {
      fndb[a>>16]=(struct fenris_fndb *) malloc(sizeof( struct fenris_fndb));
      memcpy(fndb[a>>16],&ff,sizeof( struct fenris_fndb));
    } else {
      struct fenris_fndb* dest=fndb[a>>16];
      while (dest->next) dest=dest->next;
      dest->next= (struct fenris_fndb *) malloc(sizeof (struct fenris_fndb));
      memcpy(dest->next,&ff,sizeof (struct fenris_fndb));
    }
  }     

  fclose(f);
  hide_wait_box();
//  info("The fingerprints database has been loaded");

}



/******************************
 * Handle local function call *
 ******************************/

static inline char * handle_fncall(unsigned char sig[SIGNATSIZE+4], int *count) { 
    unsigned char result[16];
    MD5_CTX kuku;
	int i;

    // I suck. TODO: parse relocs in signatures. For now,
    // just carefully exterminate anything interesting ;)
	for (i=0;i<SIGNATSIZE;i++)
     if (sig[i]== 0x08) memset(&sig[i-3], 0, 4);

    for (i=0;i<SIGNATSIZE;i++)
     if (sig[i]==0xe8) memset(&sig[i+1], 0, 4);


    MD5Init(&kuku);
    MD5Update(&kuku,sig,SIGNATSIZE);
    MD5Final(result,&kuku);

    *((int *)result) ^= *((int *) (result + 8 ));
    *((int *)(result + 4 )) ^= *((int *)(result + 12));

    return find_fprints(*((int *)result) ^= *((int *)(result + 4 )), count);
}


/**********************
 * Find a fingerprint *
 **********************/

static inline char * find_fprints(unsigned int a, int *count) {
  unsigned short sht = a & 0xffff;
  char didsay=0;
  struct fenris_fndb *cur;
  char *result;
  // Use a table to lookup first two
  // bytes instantly, then compare two remaining bytes by searching
  // a linked list. Long live optimization.

  cur=fndb[a>>16];
  *count = 0;
  result = NULL;

  while (cur) {
    if (cur->a == sht) {
		(*count)++;
      if (!didsay) {
        didsay=1;
		result = cur->name;      
      }
//      msg("%s ",cur->name);
    }
    cur=cur->next;
  }

  return result;
//  if (didsay) msg("\n");
//    else msg("# No matches for signature %08X.\n",a);

}
